/*
   Ce fichier contient un programme qui 
   calcule les dformations d'une structure
   de type treillis plan.
*/

#include <assert.h>
#include <stdio.h>
#include <math.h>
// quelques fonctions utiles 

// lire un ficher de donnees
void lire_donnees ( const char * );

// acces aux noeuds du maillage
// noeud varie de 0 a nombre_de_noeuds
int nombre_de_noeuds ();
// coordonnees du noeud inoeud
void coordonnees_du_noeud (int inoeud, double x[3]);
// retourne le nombre de deplacements fixes
int nombre_de_deplacements_fixes ();
// donne le ideplac eme deplacement fixe : le noeud fixe, la direction (0=x,1=y,2=z)
// dans laquelle le deplacement est fixe et la valeur de ce deplacement fixe
void deplacement_fixe (int ideplac, int noeud[1] , int direction[1], double valeur[1]);
// nombre de forces nodales 
int nombre_de_forces_nodales ();
// force nodale ifor d'intensite valeur appliquee au noeud 
void force_nodale (int ifor, int noeud[1] , double valeur[3]);

// acces a la connectivite i.e. aux barres
int nombre_de_barres ();
// donne la connectivite de la barre ibarre + les caracteristiques de la section + le materiau
int connectivite_de_la_barre (int ibarre, int noeuds[2], double E[1], double I[1], double A[1], double G[1]);

// manipulation d'un systeme lineaire
double ** cree_matrice_carree    (int nb_lignes);
double  * cree_vecteur           (int nb_lignes);
int     * cree_vecteur_entiers   (int nb_lignes);
void      resolution_systeme     (int nb_lignes, double **matrice, 
                               double *vecteur, double *solution); 
void      detruit_matrice_carree (double **matrice);
void      detruit_vecteur        (double *vecteur);
double ** mult_matrices_carrees (int n, double **m1, double **m2);
double * mult_matrice_vecteur (int n, double *m1, double **m2);

// raideurs
double ** matrice_raideur_poutre_2D ( double E , double A, double I, double x0[3], double x1[3] );
// A FAIRE 
double ** matrice_raideur_barre_2D ( double E , double A, double I, double x0[3], double x1[3] );
double ** matrice_raideur_barre_3D ( double E , double A, double I, double x0[3], double x1[3] );
double ** matrice_raideur_grillage ( double E , double A, double I, double G, double x0[3], double x1[3] );

// post processing
void sauve_deplacements       (int TypeAnalyse, double * solution, const char *nom_fichier);
void sauve_effortsMNT (int TypeAnalyse, double * solution, const char *nom_M, const char *nom_N,const char *nom_T);

void assemble ( int nbeq , int *lignes , double ** matrice , double ** raideur )
{
  int j,k;
  for (j=0;j<nbeq;j++)
    for (k=0;k<nbeq;k++)
      matrice [lignes[j]][lignes[k]] += raideur[j][k];
}

int main (int argc, char *argv[])
{
  int TypeAnalyse = 1; // TypeAnalyse : 1 barres 2D, 2 poutres 2D, 3 barres 3D, 4 grillage
  int NbrDdlParNoeud = 2;
  int Nombre_Equations,i,j,k,noeuds[2],ligne,direction[1],*lignes;
  double **matrice, *membre_de_droite, *solution,raideur_diagonale;
  double E[1], I[1], A[1], G[1], angle;
  double x0[3], x1[3],L,valeur[3],**raideur;
  // Le programme s'utilise comme suit
  // Treillis fichier
  // fichier est le fichier de donnees 
  
  assert (argc >= 2);
    
  if (argc == 3) 
    {
      TypeAnalyse = atoi(argv[2]);
      if (TypeAnalyse == 2) NbrDdlParNoeud = 3; 
      if (TypeAnalyse == 3) NbrDdlParNoeud = 3;
      if (TypeAnalyse == 4) NbrDdlParNoeud = 3; 
    }
  // on commence par lire les donnees
  lire_donnees (argv[1]);

  // On est en presence d'uns strcuture
  // avec nombre_de_noeuds() noeuds.  On cree un systeme d'equations
  // qui a  NbrDdlParNoeud fois le nombre de noeuds comme nombre d'inconnues

  Nombre_Equations =  NbrDdlParNoeud * nombre_de_noeuds ();
  matrice = cree_matrice_carree( Nombre_Equations );
  membre_de_droite = cree_vecteur( Nombre_Equations );
  solution = cree_vecteur( Nombre_Equations );
  lignes   = cree_vecteur_entiers (2 *  NbrDdlParNoeud);

  // on assemble ensuite les matrices elementaires
  // pour cela, on itere sur toutes les barres et 
  // on assemble la contribution de celle ci dans la
  // matrice globale.

  for (i=0;i<nombre_de_barres ();i++)
    {
      connectivite_de_la_barre (i, noeuds, E, I, A,G);
      coordonnees_du_noeud (noeuds[0],x0);
      coordonnees_du_noeud (noeuds[1],x1);

      for(j=0;j<NbrDdlParNoeud;j++)
	lignes[j] =  NbrDdlParNoeud*noeuds[0]+j;
      for(j=0;j<NbrDdlParNoeud;j++)
	lignes[j+NbrDdlParNoeud] =  NbrDdlParNoeud*noeuds[1]+j;

      if (TypeAnalyse == 1)
      	raideur =  matrice_raideur_barre_2D  ( E[0], A[0], I[0], x0, x1 );
      else if (TypeAnalyse == 2)
	raideur =  matrice_raideur_poutre_2D ( E[0] , A[0], I[0],  x0, x1 );
      else if (TypeAnalyse == 3)
	raideur =  matrice_raideur_barre_3D  ( E[0] , A[0], I[0],  x0, x1 );
      else if (TypeAnalyse == 4)
	raideur =  matrice_raideur_grillage  ( E[0] , A[0], I[0],  G[0], x0, x1 );
      assemble (2 * NbrDdlParNoeud,lignes, matrice,raideur);
      //imprime_matrice_carree(2 * NbrDdlParNoeud,raideur);
      detruit_matrice_carree(raideur);
    }
  
  // imposition des conditions aux limites par une methode de
  // penalite i.e. on ajoute un terme grand sur l'element diagonal 
  // du deplacement impose
  for (i=0;i<nombre_de_deplacements_fixes ();i++)
    {
      deplacement_fixe (i,noeuds , direction, valeur);
      if (NbrDdlParNoeud > direction[0])
	{
	  ligne =  NbrDdlParNoeud*noeuds[0] + direction[0];
	  raideur_diagonale =  matrice[ligne][ligne];
	  matrice[ligne][ligne] += (1. + raideur_diagonale) * 10e3;
	  membre_de_droite[ligne] += (1. + raideur_diagonale) * 10e3 * valeur[0];
	}
    }

  // ajout des forces nodales
  for (i=0;i<nombre_de_forces_nodales ();i++)
    {      
      force_nodale (i,noeuds , valeur);
      //      printf("force nodale %d %d %lf %lf\n",
      //	     i,noeuds[0],valeur[0],valeur[1]);
      ligne =  NbrDdlParNoeud*noeuds[0];
      membre_de_droite[ligne] += valeur[0];
      ligne =  NbrDdlParNoeud*noeuds[0] + 1;
      membre_de_droite[ligne] += valeur[1];      
    }

  for (i=0;i<Nombre_Equations;i++)if(matrice[i][i] < 1.e-5)matrice[i][i] = E[0] * A[0]/L;

  //  imprime_matrice_carree(Nombre_Equations,matrice);

  // resolution
  resolution_systeme     (Nombre_Equations, matrice, 
  			  membre_de_droite, solution); 

  //    imprime_vecteur(Nombre_Equations,solution);
  //  imprime_vecteur(Nombre_Equations,membre_de_droite);
  
  sauve_deplacements (TypeAnalyse, solution, "deplacements.pos");
  sauve_effortsMNT   (TypeAnalyse, solution, "moments.pos","normal.pos","tranchant.pos");
  // c'est fini ...
}
