#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define LGN1 1
#define TRI1 2
#define QUA1 3
#define TET1 4
#define HEX1 5
#define PRI1 6
#define PYR1 7
#define LGN2 8
#define TRI2 9
#define QUA2 10
#define TET2 11
#define HEX2 12
#define PRI2 13
#define PYR2 14
#define PNT  15 

#define TINY 1.0e-20;

static int *connectivite=0, Nbr_Nodes, nbelem, Nbr_Fix, Nbr_For;
static double **Xs,*Is,*As,*Es,**Fix,**Forces;

void ludcmp(double **a, int n, int *indx,double *d);
void lubksb(double **a, int n, int *indx, double b[]);
void invmat(double **a, double **y, int n);
void nrerror(char *error);
double *vector(int nl,int nh);
void free_vector(double *v,int nl,int nh);

double ** allocateMatrix(int nbr, int nbc)
{
  int i;
  double ** v = (double**)malloc(nbr*sizeof(double*));
  double *help = (double*)malloc(nbr*nbc*sizeof(double));
  for (i=0; i!= nbr; i++){
    v[i] = &help[i*nbc];
  }
  for (i=0;i<nbr*nbc;i++)help[i] = 0.0;
  return v;
}
void freeMatrix( double **v)
{
  if(v)
    {
      if(v[0])
	free (v[0]);
      free (v);
    }
}

void detruit_matrice_carree( double **v)
{
  freeMatrix(v);
}

void resolution_systeme(int n,double **a, double *y, double *sol)
{

  double d;
  int i,*indx;

  indx = (int *)malloc(n*sizeof(int));
  ludcmp(a,n,indx,&d);
  for(i=0; i < n; i++) sol[i] = y[i];
  lubksb(a,n,indx,sol);
  free(indx);
}

void invmat(double **a, double **y, int n)
{

  double d,*col;
  int i,j,*indx;

  col = (double *)malloc(n*sizeof(double));
  indx = (int *)malloc(n*sizeof(int));
  ludcmp(a,n,indx,&d);
  for(j=0;j<n;j++){
    for(i=0; i < n; i++) col[i] = 0.0;
    col[j]=1.0;
    lubksb(a,n,indx,col);
    for(i=0;i<n;i++) y[i][j]=col[i];
  }
  free(col);
  free(indx);
}

void ludcmp(double **a, int n, int *indx,double *d)
{
  int i, imax, j,k;
  double big,dum,sum,temp;
  double *vv;
  
  vv=vector(0,n-1);
  *d=1.0;
  for(i=0; i<n; i++){
    big=0.0;
    for(j=0;j<n; j++)
      if((temp=fabs(a[i][j])) > big) big=temp;
    if(big==0.0) {
      printf("Singular matrix in routine LUDCMP");
      big = 1.e-12;
    }
    vv[i]=1.0/big;
  }
  for(j=0; j<n;j++){
    for(i=0;i<j;i++){
      sum=a[i][j];
      for(k=0;k<i;k++) sum -= a[i][k]*a[k][j];
      a[i][j] = sum;
    }
    big=0.0;
    for(i=j; i<n;i++){
      sum=a[i][j];
      for(k=0;k<j;k++)
	sum -= a[i][k]*a[k][j];
      a[i][j]=sum;
      if((dum=vv[i]*fabs(sum)) >= big){
	big=dum;
	imax=i;
      }
    }
    if(j != imax){
      for(k=0; k <n; k++){
	dum=a[imax][k];
	a[imax][k]=a[j][k];
	a[j][k]=dum;
      }
      *d = -(*d);
      vv[imax]=vv[j];
    }
    indx[j]=imax;
    if( a[j][j] == 0.0) a[j][j] = TINY;
    if(j !=n){
      dum=1.0/(a[j][j]);
      for(i=j+1;i<n;i++) a[i][j] *= dum;
    }
  }
  free_vector(vv,0,n-1);
}



void nrerror(char *error)
{
  fprintf(stderr,"error\n");
  fprintf(stderr,"%s\n",error);
  exit(1);
}

double *vector(int nl,int nh)
{
  double *v;
  
  v=(double*)malloc((unsigned)(nh-nl+1)*sizeof(double));
  if (!v) nrerror("allocation failure in vector()");
  return (v-nl);
}

void free_vector(double *v,int nl,int dum)
{
  free((char*)(v+nl));
}

void lubksb(double **a, int n, int *indx, double b[])
{
  int i,ii=-1,ip,j;
  double sum;
  
  for(i=0; i< n; i++){
    ip=indx[i];
    sum=b[ip];
    b[ip]=b[i];
    if(ii != -1)
      for(j=ii; j <= i-1; j++) sum -= a[i][j]*b[j];
    else if (sum) ii=i;
    b[i]=sum;
  }
  for(i=n-1; i>=0; i--){
    sum=b[i];
    for(j=i+1;j<n;j++) sum -= a[i][j]*b[j];
    b[i]=sum/a[i][i];
  }
}
int nombre_de_noeuds ()
{
  return Nbr_Nodes;
}

void coordonnees_du_noeud (int noeud, double x[3])
{
  x[0] = Xs[noeud][0];
  x[1] = Xs[noeud][1];
  x[2] = Xs[noeud][2];
}

int nombre_de_barres ()
{
  return nbelem;
}
int connectivite_de_la_barre (int barre, int noeuds[2], double E[1], double I[1], double A[1], double G[1])
{
  noeuds[0] = connectivite[2*barre];
  noeuds[1] = connectivite[2*barre+1];
  E[0] = Es[barre];
  I[0] = Is[barre];
  A[0] = As[barre];
  G[0] = E[0] / (2. * (1. + 0.23));
}
double **cree_matrice_carree(int n)
{return allocateMatrix(n,n);}
double *cree_vecteur(int n)
{
  double *x;
  int i;
  x = (double*)malloc(n*sizeof(double));
  for (i=0;i<n;i++)x[i] = 0.0;
  return x;
}
int *cree_vecteur_entiers(int n)
{
  int *x;
  int i;
  x = (int*)malloc(n*sizeof(int));
  for (i=0;i<n;i++)x[i] = 0;
  return x;
}
void lire_donnees (const char *n)
{
  char String[256];
  FILE *File_GEO = fopen(n,"r");
  int *rel,nodes[8],Nbr_No;
  int  Nbr_Elements, i_Node, i_Element;
  int  Num, Type, Physical, Elementary, i, j, replace;
  double h = 0.05;

  while (1) {
    do { 
      fgets(String,sizeof(String), File_GEO) ; 
      if (feof(File_GEO))  break ;
    } while (String[0] != '$') ;  
    if (feof(File_GEO))  break ;
    if (!strncmp(&String[1], "NO", 2)) { /* $NOE or $NOD */      
      fscanf(File_GEO, "%d", &Nbr_Nodes) ;
      Xs = allocateMatrix(Nbr_Nodes,3);
      rel = (int*)malloc(25*sizeof(int)*Nbr_Nodes);//bof
      for (i = 0 ; i < Nbr_Nodes ; i++) {
	fscanf(File_GEO, "%d %lf %lf %lf", &i_Node, &Xs[i][0], &Xs[i][1], &Xs[i][2]) ;
	rel[i_Node] = i;
      }
    }
    else if (!strncmp(&String[1], "FOR", 3)) { /* $FOR */      
      fscanf(File_GEO, "%d", &Nbr_For) ;
      Forces = allocateMatrix(Nbr_For,4);
      for (i = 0 ; i < Nbr_For ; i++) {
	fscanf(File_GEO, "%lf %lf %lf %lf", &Forces[i][0], &Forces[i][1], 
	       &Forces[i][2], &Forces[i][3]) ;
	Forces[i][0] = rel[(int)	Forces[i][0] ];
      }
    }
    else if (!strncmp(&String[1], "FIX", 3)) { /* $FIX */      
      fscanf(File_GEO, "%d", &Nbr_Fix) ;
      Fix = allocateMatrix(Nbr_Fix,3);
      for (i = 0 ; i < Nbr_Fix ; i++) {
	fscanf(File_GEO, "%lf %lf %lf", &Fix[i][0], &Fix[i][1], &Fix[i][2]) ;
	Fix[i][0] = rel[(int)	Fix[i][0] ];
      }
    }
    else if (!strncmp(&String[1], "ELM", 3)) {
      fscanf(File_GEO, "%d", &Nbr_Elements) ;
      nbelem = 0;
      connectivite = (int*)malloc(2*sizeof(int)*Nbr_Elements);
      Es = (double*)malloc(sizeof(double)*Nbr_Elements);
      Is = (double*)malloc(sizeof(double)*Nbr_Elements);
      As = (double*)malloc(sizeof(double)*Nbr_Elements);
      for (i = 0 ; i < Nbr_Elements ; i++) 
	{
	  fscanf(File_GEO, "%d %d %d %d %d", 
		 &Num, &Type, &Physical, &Elementary, &Nbr_No) ;
	  for (j = 0 ; j < Nbr_No ; j++)
	    fscanf(File_GEO, "%d", &nodes[j]) ;
	  
	  switch(Type){
	  case LGN1: case LGN2:
	    Is[nbelem] = h*h*h*h /12.; 
	    As[nbelem] = h*h;
	    Es[nbelem] = 210e9;
	    connectivite[2*nbelem]   = rel[nodes[0]];
	    connectivite[2*nbelem+1] = rel[nodes[1]];
	    nbelem++;
	    break;
	  }
	}
    }
  }
}

int nombre_de_deplacements_fixes ()
{
  return Nbr_Fix;
}
void deplacement_fixe (int idf, int noeud[1] , int direction[1], double valeur[1])
{
  noeud[0] = (int)Fix[idf][0];
  direction[0] = (int)Fix[idf][1];
  valeur[0] = Fix[idf][2];
}
int nombre_de_forces_nodales ()
{
  return Nbr_For;
}
void force_nodale (int ifor, int noeud[1] , double valeur[3])
{
  noeud[0] = (int)Forces[ifor][0];
  valeur[0] = Forces[ifor][1];
  valeur[1] = Forces[ifor][2];
  valeur[2] = Forces[ifor][3];
}

double ** mult_matrices_carrees (int n, double **m1, double **m2)
{
  int I,J,K;
  double **temp;

  temp = cree_matrice_carree (n);
  for (I=0;I<n;I++)
    {
      for (J=0;J<n;J++)
	{
	  for (K=0;K<n;K++)
	    {
	      temp[I][J] +=  m1[I][K] *  m2 [K][J];
	    }
	}
    }
  return temp;
}

double * mult_matrice_vecteur (int n, double *m1, double **m2)
{
  int I,J;
  double *temp;

  temp = cree_vecteur (n);
  for (I=0;I<n;I++)
    {
      for (J=0;J<n;J++)
	{
	  temp[I] +=  m1[J] *  m2 [I][J];
	}
    }
  return temp;
}


void imprime_matrice_carree (int n, double **m1)
{
  int I,J;

  for (I=0;I<n;I++)
    {
      for (J=0;J<n;J++)
	{
	  printf("%8.5E ",m1[I][J]);
	}
      printf("\n");
    }
      printf("\n");
}
void imprime_vecteur (int n, double *m1)
{
  int I;

  for (I=0;I<n;I++)
    {
      printf("%12.5E ",m1[I]);
    }
  printf("\n");
}

void sauve_deplacements (int TypeAnalyse, double * solution, const char *nom_fichier)
{
  FILE *out = fopen (nom_fichier,"w");
  int i,noeuds[2],j;
  double G[1],E[1],A[1],I[1],x0[3],x1[3],ux1,uy1,ux2,uy2,tz1,tz2,L,angle,Ux[100],Uy[100],x[100],y[100],z[100];
  double uxp1,uyp1,uxp2,uyp2,uxp,uyp,t;
  fprintf(out,"View \"Deplacements\" {\n");
  for (i=0;i<nombre_de_barres ();i++)
    {
      connectivite_de_la_barre (i, noeuds, E, I, A,G);
      coordonnees_du_noeud (noeuds[0],x0);
      coordonnees_du_noeud (noeuds[1],x1);
      if (TypeAnalyse == 3)
	fprintf(out,"VL(%lf,%lf,%lf,%lf,%lf,%lf)  {%12.5E,%12.5E,%12.5E,%12.5E,%12.5E,%12.5E};\n",
		x0[0],x0[1],x0[2],x1[0],x1[1],x1[2],
		solution[ 3*noeuds[0]], 
		solution[ 3*noeuds[0]+1],
		solution[ 3*noeuds[0]+2],
		solution[ 3*noeuds[1]], 
		solution[ 3*noeuds[1]+1],
		solution[ 3*noeuds[1]+2]);
      else if (TypeAnalyse == 1)
	fprintf(out,"VL(%lf,%lf,%lf,%lf,%lf,%lf)  {%12.5E,%12.5E,%12.5E,%12.5E,%12.5E,%12.5E};\n",
		x0[0],x0[1],x0[2],x1[0],x1[1],x1[2],
		solution[ 2*noeuds[0]], solution[ 2*noeuds[0]+1],0.0,
		solution[ 2*noeuds[1]], solution[ 2*noeuds[1]+1],0.0);
      else if (TypeAnalyse == 2 || TypeAnalyse == 4)
	{	
	  angle = (TypeAnalyse == 4)? 0.0 :atan2 (x1[1]-x0[1],x1[0]-x0[0]);	  
	  while(angle < 0)angle += 2. * M_PI;
	  ux1 = solution[ 3*noeuds[0]];
	  uy1 = solution[ 3*noeuds[0]+1];
	  tz1 = solution[ 3*noeuds[0]+2];
	  ux2 = solution[ 3*noeuds[1]];
	  uy2 = solution[ 3*noeuds[1]+1];
	  tz2 = solution[ 3*noeuds[1]+2];

	  uxp1 = (TypeAnalyse == 4)? 0.0 : (ux1 * cos (angle) + uy1 * sin(angle));
	  uyp1 = -ux1 * sin (angle) + uy1 * cos(angle);
	  uxp2 = (TypeAnalyse == 4)? 0.0 : (ux2 * cos (angle) + uy2 * sin(angle));
	  uyp2 = -ux2 * sin (angle) + uy2 * cos(angle);

	  for (j=0;j<20;j++)
	    {
	      t = (double)j / 19.;
	      x[j] = x0[0] + t * (x1[0] - x0[0]);
	      y[j] = x0[1] + t * (x1[1] - x0[1]);
	      z[j] = x0[2] + t * (x1[2] - x0[2]);
	      uxp = uxp1 + t * (uxp2-uxp1);
	      uyp = 
		(2*t+1)*(t-1.)*(t-1.)*uyp1 +
		t*(t-1.)*(t-1.)*tz1 +
		t*t*(-2.*t+3.)*uyp2 -
		t*t*(1.-t)*tz2;
	      Ux[j] = uxp * cos (angle) - uyp * sin(angle);
	      Uy[j] = uxp * sin (angle) + uyp * cos(angle);
	    }
	  for (j=0;j<19;j++)
	    {
	      fprintf(out,"VL(%lf,%lf,%lf,%lf,%lf,%lf)  {%12.5E,%12.5E,%12.5E,%12.5E,%12.5E,%12.5E};\n",
		      x[j],y[j],z[j],x[j+1],y[j+1],z[j+1],Ux[j],Uy[j],0.0,Ux[j+1],Uy[j+1],0.0);
	    }
	}	
    }
  fprintf(out,"};\n");
  fclose(out); 
}

void sauve_effortsMNT (int TypeAnalyse, double * solution, const char *nom_M, const char *nom_N,const char *nom_T)
{
  FILE *outM = fopen (nom_M,"w");
  int i,noeuds[2],j;
  double G[1],E[1],A[1],I[1],x0[3],x1[3],ux1,uy1,ux2,uy2,tz1,tz2,L,angle,x[100],y[100],z[100],Mx[100],My[100];
  double uxp1,uyp1,uxp2,uyp2,Mx1,Mx2,My1,My2,Myp1,Myp2;
  fprintf(outM,"View \"Moments de flexion\" {\n");
  for (i=0;i<nombre_de_barres ();i++)
    {
      connectivite_de_la_barre (i, noeuds, E, I, A,G);
      coordonnees_du_noeud (noeuds[0],x0);
      coordonnees_du_noeud (noeuds[1],x1);
      if (TypeAnalyse == 2 || TypeAnalyse == 4)
	{	
	  angle = (TypeAnalyse == 4)? 0.0 : atan2 (x1[1]-x0[1],x1[0]-x0[0]);	  
	  L = sqrt ((x0[0]-x1[0]) * (x0[0]-x1[0]) +
		    (x0[1]-x1[1]) * (x0[1]-x1[1]) +
		    (x0[2]-x1[2]) * (x0[2]-x1[2]) );

	  ux1 = solution[ 3*noeuds[0]];
	  uy1 = solution[ 3*noeuds[0]+1];
	  tz1 = solution[ 3*noeuds[0]+2];
	  ux2 = solution[ 3*noeuds[1]];
	  uy2 = solution[ 3*noeuds[1]+1];
	  tz2 = solution[ 3*noeuds[1]+2];

	  uxp1 = (TypeAnalyse == 4)? 0.0 :ux1 * cos (angle) + uy1 * sin(angle);
	  uyp1 = -ux1 * sin (angle) + uy1 * cos(angle);
	  uxp2 = (TypeAnalyse == 4)? 0.0 :ux2 * cos (angle) + uy2 * sin(angle);
	  uyp2 = -ux2 * sin (angle) + uy2 * cos(angle);

	  Myp1 = E[0] * I[0] * ((-6. * uyp1 + 6 * uyp2)/(L*L) +  (-4 * tz1 - 2 * tz2)/(L));
	  Myp2 = E[0] * I[0] * ((6. * uyp1  - 6 * uyp2)/(L*L) +  (2 * tz1 + 4 * tz2)/(L));

	  Mx1 = -sin(angle) * Myp1;
	  My1 = cos(angle) * Myp1;
	  Mx2 = -sin(angle) * Myp2;
	  My2 = cos(angle) * Myp2;
	  for (j=0;j<20;j++)
	    {
	      double t = (double)j / 19.;
	      x[j] = x0[0] + t * (x1[0] - x0[0]);
	      y[j] = x0[1] + t * (x1[1] - x0[1]);
	      z[j] = x0[2] + t * (x1[2] - x0[2]);	      
	      Mx[j] = Mx1 + t * (Mx2 -Mx1); 
	      My[j] = My1 + t * (My2 -My1); 
	    }
	  for (j=0;j<19;j++)
	    {
	      fprintf(outM,"VL(%lf,%lf,%lf,%lf,%lf,%lf)  {%12.5E,%12.5E,%12.5E,%12.5E,%12.5E,%12.5E};\n",
		      x[j],y[j],z[j],x[j+1],y[j+1],z[j+1],Mx[j],My[j],0.0,Mx[j+1],My[j+1],0.0);
	    }
	  fprintf(outM,"SL(%lf,%lf,%lf,%lf,%lf,%lf)  {%12.5E,%12.5E};\n",
		  x0[0],x0[1],x0[2],x1[0],x1[1],x1[2],Myp1,Myp2);

	}	
    }
  fprintf(outM,"};\n");
  fclose(outM); 
}
